from __future__ import annotations

from typing import Union, Dict
from typing_extensions import TypeAlias

import pandas as pd

DataBlock: TypeAlias = Union[
    pd.DataFrame,
    Dict[str, Union[str, int, float]]
]



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Union

if TYPE_CHECKING:
    import pandas as pd
    from os import PathLike

from .parser import StarParser
from .writer import StarWriter
from .typing import DataBlock

if TYPE_CHECKING:
    import pandas as pd
    from os import PathLike


def read(filename: PathLike, read_n_blocks: int = None, always_dict: bool = False):
    """
    Read a star file into a pandas dataframe or dict of pandas dataframes

    default behaviour in the case of only one data block being present in the STAR file is to
    return only a dataframe, this can be changed by setting 'always_dict=True'
    """

    parser = StarParser(filename, n_blocks_to_read=read_n_blocks)
    if len(parser.data_blocks) == 1 and always_dict is False:
        return list(parser.data_blocks.values())[0]
    else:
        return parser.data_blocks


def write(
    data: Union[DataBlock, Dict[str, DataBlock], List[DataBlock]],
    filename: PathLike,
    float_format: str = '%.6f',
    sep: str = '\t',
    na_rep: str = '<NA>',
    quote_character: str = '"',
    quote_all_strings: bool = False,
    **kwargs,
):
    """Write data blocks as STAR files."""
    StarWriter(
        data,
        filename=filename,
        float_format=float_format,
        na_rep=na_rep,
        separator=sep,
        quote_character=quote_character,
        quote_all_strings=quote_all_strings,
    )



from __future__ import annotations

from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Union, Dict, List
from importlib.metadata import version
import csv

import pandas as pd

from .utils import TextBuffer
from .typing import DataBlock

if TYPE_CHECKING:
    from os import PathLike

__version__ = version("starfile")


class StarWriter:
    def __init__(
        self,
        data_blocks: Union[DataBlock, Dict[str, DataBlock], List[DataBlock]],
        filename: PathLike,
        float_format: str = '%.6f',
        separator: str = '\t',
        na_rep: str = '<NA>',
        quote_character: str = '"',
        quote_all_strings: bool = False,
    ):
        # coerce data
        self.data_blocks = self.coerce_data_blocks(data_blocks)

        # write
        self.filename = Path(filename)
        self.float_format = float_format
        self.sep = separator
        self.na_rep = na_rep
        self.quote_character = quote_character
        self.quote_all_strings = quote_all_strings
        self.buffer = TextBuffer()
        self.backup_if_file_exists()
        self.write()

    def coerce_data_blocks(
        self,
        data_blocks: Union[DataBlock, List[DataBlock], Dict[str, DataBlock]]
    ) -> Dict[str, DataBlock]:
        if isinstance(data_blocks, pd.DataFrame):
            return coerce_dataframe(data_blocks)
        elif isinstance(data_blocks, dict):
            return coerce_dict(data_blocks)
        elif isinstance(data_blocks, list):
            return coerce_list(data_blocks)
        else:
            raise ValueError(
                f'Expected \
                {pd.DataFrame}, \
                {Dict[str, pd.DataFrame]} \
                or {List[pd.DataFrame]}, \
                got {type(data_blocks)}'
            )

    def write(self):
        write_package_info(self.filename)
        write_blank_lines(self.filename, n=2)
        self.write_data_blocks()

    def write_data_blocks(self):
        for block_name, block in self.data_blocks.items():
            if isinstance(block, dict):
                write_simple_block(
                    file=self.filename,
                    block_name=block_name,
                    data=block,
                    quote_character=self.quote_character,
                    quote_all_strings=self.quote_all_strings
                )
            elif isinstance(block, pd.DataFrame):
                write_loop_block(
                    file=self.filename,
                    block_name=block_name,
                    df=block,
                    float_format=self.float_format,
                    separator=self.sep,
                    na_rep=self.na_rep,
                    quote_character=self.quote_character,
                    quote_all_strings=self.quote_all_strings
                )

    def backup_if_file_exists(self):
        if self.filename.exists():
            new_name = self.filename.name + '~'
            backup_path = self.filename.resolve().parent / new_name
            self.filename.rename(backup_path)


def coerce_dataframe(df: pd.DataFrame) -> Dict[str, DataBlock]:
    return {'': df}


def coerce_dict(
    data_blocks: Union[DataBlock, Dict[str, DataBlock]]
) -> Dict[str, DataBlock]:
    """Coerce dict into dict of data blocks."""
    # check if data is already Dict[str, DataBlock]
    for k, v in data_blocks.items():
        if type(v) in (dict, pd.DataFrame):  #
            return data_blocks
    # coerce if not
    return {'': data_blocks}


def coerce_list(data_blocks: List[DataBlock]) -> Dict[str, DataBlock]:
    """Coerces a list of DataFrames into a dict"""
    return {f'{idx}': df for idx, df in enumerate(data_blocks)}


def write_blank_lines(file: Path, n: int):
    with open(file, mode='a') as f:
        f.write('\n' * n)


def write_package_info(file: Path):
    date = datetime.now().strftime('%d/%m/%Y')
    time = datetime.now().strftime('%H:%M:%S')
    line = f'# Created by the starfile Python package (version {__version__}) at {time} on {date}'
    with open(file, mode='w+') as f:
        f.write(f'{line}\n')


def write_simple_block(
    file: Path,
    block_name: str,
    data: Dict[str, Union[str, int, float]],
    quote_character: str = '"',
    quote_all_strings: bool = False
):  
    quoted_data = {
        k: f"{quote_character}{v}{quote_character}" 
        if isinstance(v, str) and (quote_all_strings or " " in v or v == "") 
        else v
        for k, v
        in data.items()    
    }
    formatted_lines = '\n'.join(
        [
            f'_{k}\t\t\t{v}'
            for k, v
            in quoted_data.items()
        ]
    )
    with open(file, mode='a') as f:
        f.write(f'data_{block_name}\n\n')
        f.write(formatted_lines)
        f.write('\n\n\n')


def write_loop_block(
    file: Path,
    block_name: str,
    df: pd.DataFrame,
    float_format: str = '%.6f',
    separator: str = '\t',
    na_rep: str = '<NA>',
    quote_character: str = '"',
    quote_all_strings: bool = False
):
    # write header
    header_lines = [
        f'_{column_name} #{idx}'
        for idx, column_name
        in enumerate(df.columns, 1)
    ]
    with open(file, mode='a') as f:
        f.write(f'data_{block_name}\n\n')
        f.write('loop_\n')
        f.write('\n'.join(header_lines))
        f.write('\n')

    df = df.map(lambda x: f'{quote_character}{x}{quote_character}' 
                if isinstance(x, str) and (quote_all_strings or " " in x or x == "") 
                else x)

    # write data
    df.to_csv(
        path_or_buf=file,
        mode='a',
        sep=separator,
        header=False,
        index=False,
        float_format=float_format,
        na_rep=na_rep,
        quoting=csv.QUOTE_NONE
    )
    write_blank_lines(file, n=2)



from pathlib import Path

import pandas as pd

# Directories for test data
test_data_directory = Path(__file__).parent / 'data'
relion_tutorial = test_data_directory / 'relion_tutorial'

# Test files
loop_simple = test_data_directory / 'one_loop.star'
postprocess = test_data_directory / 'postprocess.star'
pipeline = test_data_directory / 'default_pipeline.star'
rln31_style = test_data_directory / 'rln3.1_data_style.star'
single_line_end_of_multiblock = test_data_directory / 'single_line_end_of_multiblock.star'
single_line_middle_of_multiblock = test_data_directory / 'single_line_middle_of_multiblock.star'
optimiser_2d = relion_tutorial / 'run_it025_optimiser_2D.star'
optimiser_3d = relion_tutorial / 'run_it025_optimiser_3D.star'
sampling_2d = relion_tutorial / 'run_it025_sampling_2D.star'
sampling_3d = relion_tutorial / 'run_it025_sampling_3D.star'
non_existant_file = test_data_directory / 'non_existant_file.star'
two_single_line_loop_blocks = test_data_directory / 'two_single_line_loop_blocks.star'
two_basic_blocks = test_data_directory / 'two_basic_blocks.star'
empty_loop = test_data_directory / 'empty_loop.star'
basic_single_quote = test_data_directory / 'basic_single_quote.star'
basic_double_quote = test_data_directory / 'basic_double_quote.star'
loop_single_quote = test_data_directory / 'loop_single_quote.star'
loop_double_quote = test_data_directory / 'loop_double_quote.star'

# Example DataFrame for testing
cars = {'Brand': ['Honda_Civic', 'Toyota_Corolla', 'Ford_Focus', 'Audi_A4'],
        'Price': [22000, 25000, 27000, 35000]
        }
test_df = pd.DataFrame.from_dict(cars)



from __future__ import annotations

from collections import deque
from io import StringIO
from linecache import getline
import shlex

import numpy as np
import pandas as pd
from pathlib import Path
from typing import TYPE_CHECKING, Union, Optional, Dict, Tuple

from starfile.typing import DataBlock

if TYPE_CHECKING:
    from os import PathLike


class StarParser:
    filename: Path
    n_lines_in_file: int
    n_blocks_to_read: int
    current_line_number: int
    data_blocks: Dict[DataBlock]

    def __init__(self, filename: PathLike, n_blocks_to_read: Optional[int] = None):
        # set filename, with path checking
        filename = Path(filename)
        if not filename.exists():
            raise FileNotFoundError(filename)
        self.filename = filename

        # setup for parsing
        self.data_blocks = {}
        self.n_lines_in_file = count_lines(self.filename)
        self.n_blocks_to_read = n_blocks_to_read

        # parse file
        self.current_line_number = 0
        self.parse_file()

    @property
    def current_line(self) -> str:
        return getline(str(self.filename), self.current_line_number).strip()

    def parse_file(self):
        while self.current_line_number <= self.n_lines_in_file:
            if len(self.data_blocks) == self.n_blocks_to_read:
                break
            elif self.current_line.startswith('data_'):
                block_name, block = self._parse_data_block()
                self.data_blocks[block_name] = block
            else:
                self.current_line_number += 1

    def _parse_data_block(self) -> Tuple[str, DataBlock]:
        # current line starts with 'data_foo'
        block_name = self.current_line[5:]  # 'data_foo' -> 'foo'
        self.current_line_number += 1

        # iterate over file,
        while self.current_line_number <= self.n_lines_in_file:
            self.current_line_number += 1
            if self.current_line.startswith('loop_'):
                return block_name, self._parse_loop_block()
            elif self.current_line.startswith('_'):  # line is simple block
                return block_name, self._parse_simple_block()

    def _parse_simple_block(self) -> Dict[str, Union[str, int, float]]:
        block = {}
        while self.current_line_number <= self.n_lines_in_file:
            if self.current_line.startswith('data'):
                break
            elif self.current_line.startswith('_'):  # '_foo bar'
                k, v = shlex.split(self.current_line)
                block[k[1:]] = numericise(v)
            self.current_line_number += 1
        return block

    def _parse_loop_block(self) -> pd.DataFrame:
        # parse loop header
        loop_column_names = deque()
        self.current_line_number += 1

        while self.current_line.startswith('_'):
            column_name = self.current_line.split()[0][1:]
            loop_column_names.append(column_name)
            self.current_line_number += 1

        # now parse the loop block data
        loop_data = deque()
        while self.current_line_number <= self.n_lines_in_file:
            if self.current_line.startswith('data_'):
                break
            loop_data.append(self.current_line)
            self.current_line_number += 1
        loop_data = '\n'.join(loop_data)
        if loop_data[-2:] != '\n':
            loop_data += '\n'

        # put string data into a dataframe
        if loop_data == '\n':
            n_cols = len(loop_column_names)
            df = pd.DataFrame(np.zeros(shape=(0, n_cols)))
        else:
            df = pd.read_csv(
                StringIO(loop_data.replace("'",'"')),
                delim_whitespace=True,
                header=None,
                comment='#',
                keep_default_na=False
            )
            df_numeric = df.apply(pd.to_numeric, errors='ignore')
            # Replace columns that are all NaN with the original string columns
            df_numeric.loc[:, df_numeric.isna().all()] = df.loc[:, df_numeric.isna().all()]
            df = df_numeric
            df.columns = loop_column_names
        return df


def count_lines(file: Path) -> int:
    with open(file, 'rb') as f:
        return sum(1 for _ in f)


def block_name_from_line(line: str) -> str:
    """'data_general' -> 'general'"""
    return line[5:]


def heading_from_line(line: str) -> str:
    """'_rlnSpectralIndex #1' -> 'rlnSpectralIndex'."""
    return line.split()[0][1:]


def numericise(value: str) -> Union[str, int, float]:
    try:
        # Try to convert the string value to an integer
        value = int(value)
    except ValueError:
        try:
            # If it's not an integer, try to convert it to a float
            value = float(value)
        except ValueError:
            # If it's not a float either, leave it as a string
            value = value
    return value



from os.path import join as join_path
from tempfile import TemporaryDirectory
import time
import math 

import pandas as pd
import pytest

from starfile.parser import StarParser
from starfile.writer import StarWriter

from .constants import loop_simple, postprocess, test_data_directory, test_df
from .utils import generate_large_star_file, remove_large_star_file

def test_write_simple_block():
    s = StarParser(postprocess)
    output_file = test_data_directory / 'basic_block.star'
    StarWriter(s.data_blocks, output_file)
    assert output_file.exists()


def test_write_loop():
    s = StarParser(loop_simple)
    output_file = test_data_directory / 'loop_block.star'
    StarWriter(s.data_blocks, output_file)
    assert output_file.exists()


def test_write_multiblock():
    s = StarParser(postprocess)
    output_file = test_data_directory / 'multiblock.star'
    StarWriter(s.data_blocks, output_file)
    assert output_file.exists()


def test_from_single_dataframe():
    output_file = test_data_directory / 'from_df.star'

    StarWriter(test_df, output_file)
    assert output_file.exists()

    s = StarParser(output_file)


def test_create_from_dataframes():
    dfs = [test_df, test_df]

    output_file = test_data_directory / 'from_list.star'
    StarWriter(dfs, output_file)
    assert output_file.exists()

    s = StarParser(output_file)
    assert len(s.data_blocks) == 2


def test_can_write_non_zero_indexed_one_row_dataframe():
    # see PR #13 - https://github.com/alisterburt/starfile/pull/13
    df = pd.DataFrame([[1, 2, 3]], columns=["A", "B", "C"])
    df.index += 1

    with TemporaryDirectory() as directory:
        filename = join_path(directory, "test.star")
        StarWriter(df, filename)
        with open(filename) as output_file:
            output = output_file.read()

    expected = (
        "_A #1\n"
        "_B #2\n"
        "_C #3\n"
        "1\t2\t3"
    )
    assert (expected in output)


@pytest.mark.parametrize("quote_character, quote_all_strings, num_quotes", 
                         [('"', False, 6),
                          ('"', True, 8),
                          ("'", False, 6),
                          ("'", True, 8)
                         ])
def test_string_quoting_loop_datablock(quote_character, quote_all_strings, num_quotes, tmp_path):
    df = pd.DataFrame([[1,"nospace", "String with space", " ", ""]],
                       columns=["a_number","string_without_space", "string_space", "just_space", "empty_string"])

    filename = tmp_path / "test.star"
    StarWriter(df, filename, quote_character=quote_character, quote_all_strings=quote_all_strings)
    
    # Test for the appropriate number of quotes
    with open(filename) as f:
        star_content = f.read()
        assert star_content.count(quote_character) == num_quotes

    s = StarParser(filename)
    assert df.equals(s.data_blocks[""])

def test_writing_speed():
    start = time.time()
    generate_large_star_file()
    end = time.time()
    remove_large_star_file()

    # Check that execution takes less than a second
    assert end - start < 1

@pytest.mark.parametrize("quote_character, quote_all_strings, num_quotes", 
                         [('"', False, 6),
                          ('"', True, 8),
                          ("'", False, 6),
                          ("'", True, 8)
                         ])
def test_string_quoting_simple_datablock(quote_character, quote_all_strings,num_quotes, tmp_path):
    o = {
        "a_number": 1,
        "string_without_space": "nospace",
        "string_space": "String with space",
        "just_space": " ",
        "empty_string": ""
    }

    filename = tmp_path / "test.star"
    StarWriter(o, filename, quote_character=quote_character, quote_all_strings=quote_all_strings)
    
    # Test for the appropriate number of quotes
    with open(filename) as f:
        star_content = f.read()
        assert star_content.count(quote_character) == num_quotes

    s = StarParser(filename)
    assert o == s.data_blocks[""]



import time

import pandas as pd
import numpy as np
import pytest

from starfile.parser import StarParser
from .constants import (
    loop_simple,
    postprocess,
    pipeline,
    rln31_style,
    optimiser_2d,
    optimiser_3d,
    sampling_2d,
    sampling_3d,
    single_line_middle_of_multiblock,
    single_line_end_of_multiblock,
    non_existant_file,
    two_single_line_loop_blocks,
    two_basic_blocks,
    empty_loop,
    basic_single_quote,
    basic_double_quote,
    loop_single_quote,
    loop_double_quote,
)
from .utils import generate_large_star_file, remove_large_star_file, million_row_file


def test_instantiation():
    """
    Tests instantiation of the StarFile class
    """
    # instantiation with file which exists
    s = StarParser(loop_simple)

    # instantiation with non-existant file should fail
    assert non_existant_file.exists() is False
    with pytest.raises(FileNotFoundError):
        s = StarParser(non_existant_file)


def test_read_loop_block():
    """
    Check that loop block is parsed correctly, data has the correct shape
    """
    parser = StarParser(loop_simple)

    # Check that only one object is present
    assert len(parser.data_blocks) == 1

    # get dataframe
    df = list(parser.data_blocks.values())[0]
    assert isinstance(df, pd.DataFrame)

    # Check shape of dataframe
    assert df.shape == (16, 12)

    # check columns
    expected_columns = [
        'rlnCoordinateX',
        'rlnCoordinateY',
        'rlnCoordinateZ',
        'rlnMicrographName',
        'rlnMagnification',
        'rlnDetectorPixelSize',
        'rlnCtfMaxResolution',
        'rlnImageName',
        'rlnCtfImage',
        'rlnAngleRot',
        'rlnAngleTilt',
        'rlnAnglePsi',
    ]
    assert all(df.columns == expected_columns)


def test_read_multiblock_file():
    """
    Check that multiblock STAR files such as postprocess RELION files
    parse properly
    """
    parser = StarParser(postprocess)
    assert len(parser.data_blocks) == 3

    assert 'general' in parser.data_blocks
    assert isinstance(parser.data_blocks['general'], dict)
    assert len(parser.data_blocks['general']) == 6
    columns = list(parser.data_blocks['general'].keys())
    expected_columns = [
        'rlnFinalResolution',
        'rlnBfactorUsedForSharpening',
        'rlnUnfilteredMapHalf1',
        'rlnUnfilteredMapHalf2',
        'rlnMaskName',
        'rlnRandomiseFrom',
    ]
    assert columns == expected_columns

    assert 'fsc' in parser.data_blocks
    assert isinstance(parser.data_blocks['fsc'], pd.DataFrame)
    assert parser.data_blocks['fsc'].shape == (49, 7)

    assert 'guinier' in parser.data_blocks
    assert isinstance(parser.data_blocks['guinier'], pd.DataFrame)
    assert parser.data_blocks['guinier'].shape == (49, 3)


def test_read_pipeline():
    """
    Check that a pipeline.star file is parsed correctly
    """
    parser = StarParser(pipeline)

    # Check that data match file contents
    assert isinstance(parser.data_blocks['pipeline_general'], dict)
    assert parser.data_blocks['pipeline_processes'].shape == (31, 4)
    assert parser.data_blocks['pipeline_nodes'].shape == (74, 2)
    assert parser.data_blocks['pipeline_input_edges'].shape == (48, 2)
    assert parser.data_blocks['pipeline_output_edges'].shape == (72, 2)


def test_read_rln31():
    """
    Check that reading of RELION 3.1 style star files works properly
    """
    s = StarParser(rln31_style)

    for key, df in s.data_blocks.items():
        assert isinstance(df, pd.DataFrame)

    assert isinstance(s.data_blocks['block_1'], pd.DataFrame)
    assert isinstance(s.data_blocks['block_2'], pd.DataFrame)
    assert isinstance(s.data_blocks['block_3'], pd.DataFrame)


def test_read_n_blocks():
    """
    Check that passing read_n_blocks allows reading of only a specified
    number of data blocks from a star file
    """
    # test 1 block
    s = StarParser(postprocess, n_blocks_to_read=1)
    assert len(s.data_blocks) == 1

    # test 2 blocks
    s = StarParser(postprocess, n_blocks_to_read=2)
    assert len(s.data_blocks) == 2


def test_single_line_middle_of_multiblock():
    s = StarParser(single_line_middle_of_multiblock)
    assert len(s.data_blocks) == 2


def test_single_line_end_of_multiblock():
    s = StarParser(single_line_end_of_multiblock)
    assert len(s.data_blocks) == 2

    # iterate over dataframes, checking keys, names and shapes
    for idx, (key, df) in enumerate(s.data_blocks.items()):
        if idx == 0:
            assert key == 'block_1'
            assert df.shape == (2, 5)
        if idx == 1:
            assert key == 'block_2'
            assert df.shape == (1, 5)


def test_read_optimiser_2d():
    parser = StarParser(optimiser_2d)
    assert len(parser.data_blocks) == 1
    assert len(parser.data_blocks['optimiser_general']) == 84


def test_read_optimiser_3d():
    parser = StarParser(optimiser_3d)
    assert len(parser.data_blocks) == 1
    assert len(parser.data_blocks['optimiser_general']) == 84


def test_read_sampling_2d():
    parser = StarParser(sampling_2d)
    assert len(parser.data_blocks) == 1
    assert len(parser.data_blocks['sampling_general']) == 12


def test_read_sampling_3d():
    parser = StarParser(sampling_3d)
    assert len(parser.data_blocks) == 2
    assert len(parser.data_blocks['sampling_general']) == 15
    assert parser.data_blocks['sampling_directions'].shape == (192, 2)


def test_parsing_speed():
    generate_large_star_file()
    start = time.time()
    s = StarParser(million_row_file)
    end = time.time()
    remove_large_star_file()

    # Check that execution takes less than a second
    assert end - start < 1


def test_two_single_line_loop_blocks():
    parser = StarParser(two_single_line_loop_blocks)
    assert len(parser.data_blocks) == 2

    np.testing.assert_array_equal(
        parser.data_blocks['block_0'].columns, [f'val{i}' for i in (1, 2, 3)]
    )
    assert parser.data_blocks['block_0'].shape == (1, 3)

    np.testing.assert_array_equal(
        parser.data_blocks['block_1'].columns, [f'col{i}' for i in (1, 2, 3)]
    )
    assert parser.data_blocks['block_1'].shape == (1, 3)


def test_two_basic_blocks():
    parser = StarParser(two_basic_blocks)
    assert len(parser.data_blocks) == 2
    assert 'block_0' in parser.data_blocks
    b0 = parser.data_blocks['block_0']
    assert b0 == {
        'val1': 1.0,
        'val2': 2.0,
        'val3': 3.0,
    }
    assert 'block_1' in parser.data_blocks
    b1 = parser.data_blocks['block_1']
    assert b1 == {
        'col1': 'A',
        'col2': 'B',
        'col3': 'C',
    }


def test_empty_loop_block():
    """Parsing an empty loop block should return an empty dataframe."""
    parser = StarParser(empty_loop)
    assert len(parser.data_blocks) == 1



@pytest.mark.parametrize("quote_character, filename", [("'",basic_single_quote), 
                                                 ('"',basic_double_quote), 
                                                 ])
def test_quote_basic(quote_character,filename):
    import math
    parser = StarParser(filename)
    assert len(parser.data_blocks) == 1
    assert parser.data_blocks['']['no_quote_string'] == "noquote"
    assert parser.data_blocks['']['quote_string'] == "quote string"
    assert parser.data_blocks['']['whitespace_string'] == " "
    assert parser.data_blocks['']['empty_string'] == ""

@pytest.mark.parametrize("quote_character, filename", [("'",loop_single_quote), 
                                                 ('"',loop_double_quote), 
                                                 ])
def test_quote_loop(quote_character,filename):
    import math
    parser = StarParser(filename)
    assert len(parser.data_blocks) == 1
    assert parser.data_blocks[''].loc[0,'no_quote_string'] == "noquote"
    assert parser.data_blocks[''].loc[0,'quote_string'] == "quote string"
    assert parser.data_blocks[''].loc[0,'whitespace_string'] == " "
    assert parser.data_blocks[''].loc[0,'empty_string'] == ""

    assert parser.data_blocks[''].dtypes['number_and_string'] == object
    assert parser.data_blocks[''].dtypes['number_and_empty'] == 'float64'
    assert parser.data_blocks[''].dtypes['number'] == 'float64'
    assert parser.data_blocks[''].dtypes['empty_string_and_normal_string'] == object

    assert math.isnan(parser.data_blocks[''].loc[1,'number_and_empty'])
    assert parser.data_blocks[''].loc[0,'empty_string_and_normal_string'] == ''



from __future__ import annotations

from collections import deque
from linecache import getline, checkcache
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from os import PathLike


class TextBuffer:
    def __init__(self):
        self.buffer = deque()

    @property
    def is_empty(self) -> bool:
        if len(self.buffer) == 0:
            return True
        elif len(self.buffer) <= 100:  # arbitrary, avoid iterating large buffer
            return all([item.strip() == '' for item in self.buffer])
        else:
            return False

    def clear(self):
        self.buffer = deque()

    def add_line(self, line):
        self.buffer.append(f'{line}')

    def add_comment(self, line):
        self.add_line(f'# {line}')

    def add_blank_line(self):
        self.add_line('')

    def add_blank_lines(self, n: int):
        for i in range(n):
            self.add_blank_line()

    def as_str(self):
        buffer_with_newlines = '\n'.join(self.buffer)
        if buffer_with_newlines[-2:] != '\n':
            buffer_with_newlines = f'{buffer_with_newlines}\n'
        return buffer_with_newlines

    def write_to_disk(self, filename: PathLike, mode: str):
        with open(filename, mode) as file:
            file.write(self.as_str())

    def write_as_new_file(self, filename: PathLike):
        self.write_to_disk(filename, 'w+')

    def write_as_new_file_and_clear(self, filename: PathLike):
        self.write_as_new_file(filename)
        self.clear()

    def append_to_file(self, filename: PathLike):
        self.write_to_disk(filename, 'a')

    def append_to_file_and_clear(self, filename: PathLike):
        self.append_to_file(filename)
        self.clear()


class TextCrawler:
    def __init__(self, filename: PathLike):
        self.filename = filename
        self.current_line_number = 0
        checkcache(str(self.filename))

    def count_lines(self):
        with open(self.filename, 'rb') as f:
            return sum(1 for line in f)

    @property
    def current_line_number(self):
        return self._current_line_number

    @current_line_number.setter
    def current_line_number(self, n: int):
        self._current_line_number = n

    @property
    def current_line(self):
        return self.get_line(self.current_line_number)

    @current_line_number.setter
    def current_line_number(self, line_number: int):
        self._current_line_number = line_number

    def get_line(self, line_number: int):
        return getline(str(self.filename), line_number).strip()

    def increment_line_number(self):
        self._current_line_number += 1
